class SnowflakeDMFHandler:
    def create_dmf(
        self, dmf_name: str, dmf_args: str, dmf_comment: str, dmf_sql: str
    ) -> str:
        return f"""
            CREATE or REPLACE DATA METRIC FUNCTION
            {dmf_name} ({dmf_args})
            RETURNS NUMBER
            COMMENT = '{dmf_comment}'
            AS
            $$
            {dmf_sql}
            $$;
            """

    def add_dmf_to_table(
        self, dmf_name: str, dmf_col_args: str, dmf_schedule: str, table_identifier: str
    ) -> str:
        return f"""
            ALTER TABLE {table_identifier} SET DATA_METRIC_SCHEDULE = '{dmf_schedule}';
            ALTER TABLE {table_identifier} ADD DATA METRIC FUNCTION {dmf_name} ON ({dmf_col_args});
            """



import logging
from datetime import datetime
from json import JSONDecodeError
from typing import Dict, List, Optional, Tuple
from urllib.parse import urlparse

import dateutil.parser
import requests
from pydantic import Field, root_validator

from datahub.ingestion.api.decorators import (
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import (
    CapabilityReport,
    SourceCapability,
    TestableSource,
    TestConnectionReport,
)
from datahub.ingestion.source.dbt.dbt_common import (
    DBTColumn,
    DBTCommonConfig,
    DBTNode,
    DBTSourceBase,
)
from datahub.ingestion.source.dbt.dbt_tests import DBTTest, DBTTestResult

logger = logging.getLogger(__name__)


class DBTCloudConfig(DBTCommonConfig):
    access_url: str = Field(
        description="The base URL of the dbt Cloud instance to use. This should be the URL you use to access the dbt Cloud UI. It should include the scheme (http/https) and not include a trailing slash. See the access url for your dbt Cloud region here: https://docs.getdbt.com/docs/cloud/about-cloud/regions-ip-addresses",
        default="https://cloud.getdbt.com",
    )

    metadata_endpoint: str = Field(
        default="https://metadata.cloud.getdbt.com/graphql",
        description="The dbt Cloud metadata API endpoint. If not provided, we will try to infer it from the access_url.",
    )

    token: str = Field(
        description="The API token to use to authenticate with DBT Cloud.",
    )

    account_id: int = Field(
        description="The DBT Cloud account ID to use.",
    )
    project_id: int = Field(
        description="The dbt Cloud project ID to use.",
    )

    job_id: int = Field(
        description="The ID of the job to ingest metadata from.",
    )
    run_id: Optional[int] = Field(
        None,
        description="The ID of the run to ingest metadata from. If not specified, we'll default to the latest run.",
    )

    @root_validator(pre=True)
    def set_metadata_endpoint(cls, values: dict) -> dict:
        if values.get("access_url") and not values.get("metadata_endpoint"):
            metadata_endpoint = infer_metadata_endpoint(values["access_url"])
            if metadata_endpoint is None:
                raise ValueError(
                    "Unable to infer the metadata endpoint from the access URL. Please provide a metadata endpoint."
                )
            logger.info(f"Inferred metadata endpoint: {metadata_endpoint}")
            values["metadata_endpoint"] = metadata_endpoint
        return values


def infer_metadata_endpoint(access_url: str) -> Optional[str]:
    """Infer the dbt metadata endpoint from the access URL.

    See https://docs.getdbt.com/docs/cloud/about-cloud/access-regions-ip-addresses#api-access-urls
    and https://docs.getdbt.com/docs/dbt-cloud-apis/discovery-querying#discovery-api-endpoints
    for more information.

    Args:
        access_url: The dbt Cloud access URL. This is the URL of the dbt Cloud UI.

    Returns:
        The metadata endpoint, or None if it couldn't be inferred.

    Examples:
        # Standard multi-tenant deployments.
        >>> infer_metadata_endpoint("https://cloud.getdbt.com")
        'https://metadata.cloud.getdbt.com/graphql'

        >>> infer_metadata_endpoint("https://au.dbt.com")
        'https://metadata.au.dbt.com/graphql'

        >>> infer_metadata_endpoint("https://emea.dbt.com")
        'https://metadata.emea.dbt.com/graphql'

        # Cell-based deployment.
        >>> infer_metadata_endpoint("https://prefix.us1.dbt.com")
        'https://prefix.metadata.us1.dbt.com/graphql'

        # Test with an "internal" URL.
        >>> infer_metadata_endpoint("http://dbt.corp.internal")
        'http://metadata.dbt.corp.internal/graphql'
    """

    try:
        parsed_uri = urlparse(access_url)
        assert parsed_uri.scheme is not None
        assert parsed_uri.hostname is not None
    except Exception as e:
        logger.debug(f"Unable to parse access URL {access_url}: {e}", exc_info=e)
        return None

    if parsed_uri.hostname.endswith(".getdbt.com") or parsed_uri.hostname in {
        # Two special cases of multi-tenant deployments that use the dbt.com domain
        # instead of getdbt.com.
        "au.dbt.com",
        "emea.dbt.com",
    }:
        return f"{parsed_uri.scheme}://metadata.{parsed_uri.netloc}/graphql"
    elif parsed_uri.hostname.endswith(".dbt.com"):
        # For cell-based deployments.
        # prefix.region.dbt.com -> prefix.metadata.region.dbt.com
        hostname_parts = parsed_uri.hostname.split(".", maxsplit=1)
        return f"{parsed_uri.scheme}://{hostname_parts[0]}.metadata.{hostname_parts[1]}/graphql"
    else:
        # The self-hosted variants also have the metadata. prefix.
        return f"{parsed_uri.scheme}://metadata.{parsed_uri.netloc}/graphql"


_DBT_GRAPHQL_COMMON_FIELDS = """
  runId
  accountId
  projectId
  environmentId
  jobId
  resourceType
  uniqueId
  name
  description
  meta
  dbtVersion
  tags
"""

_DBT_GRAPHQL_NODE_COMMON_FIELDS = """
  database
  schema
  type
  owner
  comment

  columns {
    name
    index
    type
    comment
    description
    tags
    meta
  }

  # We don't currently support this field, but should in the future.
  #stats {
  #  id
  #  label
  #  description
  #  include
  #  value
  #}
"""

_DBT_GRAPHQL_MODEL_SEED_SNAPSHOT_FIELDS = """
  packageName
  alias
  error
  status
  skip
  rawSql
  rawCode
  compiledSql
  compiledCode
"""

_DBT_FIELDS_BY_TYPE = {
    "models": f"""
    { _DBT_GRAPHQL_COMMON_FIELDS }
    { _DBT_GRAPHQL_NODE_COMMON_FIELDS }
    { _DBT_GRAPHQL_MODEL_SEED_SNAPSHOT_FIELDS }
    dependsOn
    materializedType
""",
    "seeds": f"""
    { _DBT_GRAPHQL_COMMON_FIELDS }
    { _DBT_GRAPHQL_NODE_COMMON_FIELDS }
    { _DBT_GRAPHQL_MODEL_SEED_SNAPSHOT_FIELDS }
""",
    "sources": f"""
    { _DBT_GRAPHQL_COMMON_FIELDS }
    { _DBT_GRAPHQL_NODE_COMMON_FIELDS }
    identifier
    sourceName
    sourceDescription
    maxLoadedAt
    snapshottedAt
    state
    freshnessChecked
    loader
""",
    "snapshots": f"""
    { _DBT_GRAPHQL_COMMON_FIELDS }
    { _DBT_GRAPHQL_NODE_COMMON_FIELDS }
    { _DBT_GRAPHQL_MODEL_SEED_SNAPSHOT_FIELDS }
    parentsSources {{
      uniqueId
    }}
    parentsModels {{
      uniqueId
    }}
""",
    "tests": f"""
    { _DBT_GRAPHQL_COMMON_FIELDS }
    state
    columnName
    status
    error
    dependsOn
    fail
    warn
    skip
    rawSql
    rawCode
    compiledSql
    compiledCode
""",
    # Currently unsupported dbt node types:
    # - metrics
    # - snapshots
    # - exposures
}

_DBT_GRAPHQL_QUERY = """
query DatahubMetadataQuery_{type}($jobId: BigInt!, $runId: BigInt) {{
  job(id: $jobId, runId: $runId) {{
    {type} {{
{fields}
    }}
  }}
}}
"""


@platform_name("dbt")
@config_class(DBTCloudConfig)
@support_status(SupportStatus.INCUBATING)
@capability(SourceCapability.DELETION_DETECTION, "Enabled via stateful ingestion")
@capability(SourceCapability.LINEAGE_COARSE, "Enabled by default")
class DBTCloudSource(DBTSourceBase, TestableSource):
    config: DBTCloudConfig

    @classmethod
    def create(cls, config_dict, ctx):
        config = DBTCloudConfig.parse_obj(config_dict)
        return cls(config, ctx, "dbt")

    @staticmethod
    def test_connection(config_dict: dict) -> TestConnectionReport:
        test_report = TestConnectionReport()
        try:
            source_config = DBTCloudConfig.parse_obj_allow_extras(config_dict)
            DBTCloudSource._send_graphql_query(
                metadata_endpoint=source_config.metadata_endpoint,
                token=source_config.token,
                query=_DBT_GRAPHQL_QUERY.format(type="tests", fields="jobId"),
                variables={
                    "jobId": source_config.job_id,
                    "runId": source_config.run_id,
                },
            )
            test_report.basic_connectivity = CapabilityReport(capable=True)
        except Exception as e:
            test_report.basic_connectivity = CapabilityReport(
                capable=False, failure_reason=str(e)
            )
        return test_report

    @staticmethod
    def _send_graphql_query(
        metadata_endpoint: str, token: str, query: str, variables: Dict
    ) -> Dict:
        logger.debug(f"Sending GraphQL query to dbt Cloud: {query}")
        response = requests.post(
            metadata_endpoint,
            json={
                "query": query,
                "variables": variables,
            },
            headers={
                "Authorization": f"Bearer {token}",
                "X-dbt-partner-source": "acryldatahub",
            },
        )

        try:
            res = response.json()
            if "errors" in res:
                raise ValueError(
                    f'Unable to fetch metadata from dbt Cloud: {res["errors"]}'
                )
            data = res["data"]
        except JSONDecodeError as e:
            response.raise_for_status()
            raise e

        return data

    def load_nodes(self) -> Tuple[List[DBTNode], Dict[str, Optional[str]]]:
        # TODO: In dbt Cloud, commands are scheduled as part of jobs, where
        # each job can have multiple runs. We currently only fully support
        # jobs that do a full / mostly full build of the project, and will
        # print out warnings if data is missing. In the future, it'd be nice
        # to automatically detect all jobs that are part of a project and combine
        # their metadata together.
        # Additionally, we'd like to model dbt Cloud jobs/runs in DataHub
        # as DataProcesses or DataJobs.

        raw_nodes = []
        for node_type, fields in _DBT_FIELDS_BY_TYPE.items():
            logger.info(f"Fetching {node_type} from dbt Cloud")
            data = self._send_graphql_query(
                metadata_endpoint=self.config.metadata_endpoint,
                token=self.config.token,
                query=_DBT_GRAPHQL_QUERY.format(type=node_type, fields=fields),
                variables={
                    "jobId": self.config.job_id,
                    "runId": self.config.run_id,
                },
            )

            raw_nodes.extend(data["job"][node_type])

        nodes = [self._parse_into_dbt_node(node) for node in raw_nodes]

        additional_metadata: Dict[str, Optional[str]] = {
            "project_id": str(self.config.project_id),
            "account_id": str(self.config.account_id),
            "job_id": str(self.config.job_id),
        }

        return nodes, additional_metadata

    def _parse_into_dbt_node(self, node: Dict) -> DBTNode:
        key = node["uniqueId"]

        name = node["name"]
        if self.config.use_identifiers and node.get("identifier"):
            name = node["identifier"]
        if node["resourceType"] != "test" and node.get("alias"):
            name = node["alias"]

        comment = node.get("comment", "")
        description = node["description"]
        if node.get("sourceDescription"):
            description = node["sourceDescription"]

        if node["resourceType"] == "model":
            materialization = node["materializedType"]
        elif node["resourceType"] == "snapshot":
            materialization = "snapshot"
        else:
            materialization = None

        if node["resourceType"] == "snapshot":
            upstream_nodes = [
                obj["uniqueId"]
                for obj in [
                    *node.get("parentsModels", []),
                    *node.get("parentsSources", []),
                ]
            ]
        else:
            upstream_nodes = node.get("dependsOn", [])

        catalog_type = node.get("type")

        meta = node["meta"]

        # The dbt owner field is set to the db user, and not the meta property.
        # owner = node.get("owner")
        owner = meta.get("owner")

        tags = node["tags"]
        tags = [self.config.tag_prefix + tag for tag in tags]

        if node["resourceType"] in {"model", "seed", "snapshot"}:
            status = node["status"]
            if status is None and materialization != "ephemeral":
                self.report.report_warning(
                    key, "node is missing a status, schema metadata will be incomplete"
                )

            # The code fields are new in dbt 1.3, and replace the sql ones.
            raw_code = node["rawCode"] or node["rawSql"]
            compiled_code = node["compiledCode"] or node["compiledSql"]
        else:
            raw_code = None
            compiled_code = None

        max_loaded_at = None
        if node["resourceType"] == "source":
            max_loaded_at_str = node["maxLoadedAt"]
            if max_loaded_at_str:
                max_loaded_at = dateutil.parser.parse(max_loaded_at_str)

                # For missing data, dbt returns 0001-01-01T00:00:00.000Z.
                if max_loaded_at.year <= 1:
                    max_loaded_at = None

        columns = []
        if "columns" in node and node["columns"] is not None:
            # columns will be empty for ephemeral models
            columns = list(
                sorted(
                    [self._parse_into_dbt_column(column) for column in node["columns"]],
                    key=lambda c: c.index,
                )
            )

        test_info = None
        test_result = None
        if node["resourceType"] == "test":
            qualified_test_name = name

            # The qualified test name should be the test name from the dbt project.
            # It can be simple (e.g. 'unique') or prefixed (e.g. 'dbt_expectations.expect_column_values_to_not_be_null').
            # We attempt to guess the test name based on the macros used.
            for dependency in node["dependsOn"]:
                # An example dependsOn list could be:
                #     ['model.sample_dbt.monthly_billing_with_cust', 'macro.dbt.test_not_null', 'macro.dbt.get_where_subquery']
                # In that case, the test should be `not_null`.

                if dependency.startswith("macro."):
                    _, macro = dependency.split(".", 1)
                    if macro.startswith("dbt."):
                        if not macro.startswith("dbt.test_"):
                            continue
                        macro = macro[len("dbt.test_") :]

                    qualified_test_name = macro
                    break

            test_info = DBTTest(
                qualified_test_name=qualified_test_name,
                column_name=node["columnName"],
                kw_args={},  # TODO: dbt Cloud doesn't expose the args.
            )
            if not node["skip"]:
                test_result = DBTTestResult(
                    invocation_id=f"job{node['jobId']}-run{node['runId']}",
                    execution_time=datetime.now(),  # TODO: dbt Cloud doesn't expose this.
                    status=node["status"],
                    native_results={
                        key: str(node[key])
                        for key in {
                            "columnName",
                            "error",
                            "fail",
                            "warn",
                            "skip",
                            "state",
                            "status",
                        }
                    },
                )

        return DBTNode(
            dbt_name=key,
            # TODO: Get the dbt adapter natively.
            dbt_adapter=self.config.target_platform,
            dbt_package_name=node.get("packageName"),
            database=node.get("database"),
            schema=node.get("schema"),
            name=name,
            alias=node.get("alias"),
            dbt_file_path=None,  # TODO: Get this from the dbt API.
            node_type=node["resourceType"],
            max_loaded_at=max_loaded_at,
            comment=comment,
            description=description,
            upstream_nodes=upstream_nodes,
            materialization=materialization,
            catalog_type=catalog_type,
            missing_from_catalog=False,  # This doesn't really apply to dbt Cloud.
            meta=meta,
            query_tag={},  # TODO: Get this from the dbt API.
            tags=tags,
            owner=owner,
            language="sql",  # TODO: dbt Cloud doesn't surface this
            raw_code=raw_code,
            compiled_code=compiled_code,
            columns=columns,
            test_info=test_info,
            test_results=[test_result] if test_result else [],
            model_performances=[],  # TODO: support model performance with dbt Cloud
        )

    def _parse_into_dbt_column(
        self,
        column: Dict,
    ) -> DBTColumn:
        return DBTColumn(
            name=column["name"],
            comment=column.get("comment", ""),
            description=column["description"],
            # For some reason, the index sometimes comes back as None from the dbt Cloud API.
            # In that case, we just assume that the column is at the end of the table by
            # assigning it a very large index.
            index=column["index"] if column["index"] is not None else 10**6,
            data_type=column["type"],
            meta=column["meta"],
            tags=column["tags"],
        )

    def get_external_url(self, node: DBTNode) -> Optional[str]:
        # TODO: Once dbt Cloud supports deep linking to specific files, we can use that.
        return f"{self.config.access_url}/develop/{self.config.account_id}/projects/{self.config.project_id}"



import logging
from typing import Optional, Tuple

from datahub.ingestion.source.powerbi.m_query import resolver

logger = logging.getLogger(__name__)


def validate_parse_tree(
    expression: str, native_query_enabled: bool = True
) -> Tuple[bool, Optional[str]]:
    """
    :param expression: M-Query expression to check if supported data-function is present in expression
    :param native_query_enabled: Whether user want to extract lineage from native query
    :return: True or False.
    """
    function_names = [fun.value for fun in resolver.FunctionName]
    if not any(fun in expression for fun in function_names):
        return False, "DataAccess function is not present in M-Query expression."

    if native_query_enabled is False:
        if resolver.FunctionName.NATIVE_QUERY.value in function_names:
            return (
                False,
                "Lineage extraction from native query is disabled. Enable native_query_parsing in recipe",
            )

    return True, None



from typing import Optional, Tuple

import sqlglot

from datahub.sql_parsing.sql_parsing_common import QueryType, QueryTypeProps
from datahub.sql_parsing.sqlglot_utils import (
    DialectOrStr,
    get_dialect,
    is_dialect_instance,
)


def _is_temp_table(table: sqlglot.exp.Table, dialect: sqlglot.Dialect) -> bool:
    identifier: sqlglot.exp.Identifier = table.this

    return identifier.args.get("temporary") or (
        # These dialects use # as a prefix for temp tables.
        is_dialect_instance(
            dialect,
            [
                "redshift",
                "mssql",
                # sybase is another one, but we don't support that dialect yet.
            ],
        )
        and identifier.name.startswith("#")
    )


def _get_create_type_from_kind(kind: Optional[str]) -> QueryType:
    if kind and "TABLE" in kind:
        return QueryType.CREATE_TABLE_AS_SELECT
    elif kind and "VIEW" in kind:
        return QueryType.CREATE_VIEW
    else:
        return QueryType.CREATE_OTHER


def get_query_type_of_sql(
    expression: sqlglot.exp.Expression, dialect: DialectOrStr
) -> Tuple[QueryType, QueryTypeProps]:
    dialect = get_dialect(dialect)
    query_type_props: QueryTypeProps = {}

    # For creates, we need to look at the inner expression.
    if isinstance(expression, sqlglot.exp.Create):
        if is_create_table_ddl(expression):
            return QueryType.CREATE_DDL, query_type_props

        kind = expression.args.get("kind")
        if kind:
            kind = kind.upper()
            query_type_props["kind"] = kind

        target = expression.this
        if any(
            isinstance(prop, sqlglot.exp.TemporaryProperty)
            for prop in (expression.args.get("properties") or [])
        ) or _is_temp_table(target, dialect=dialect):
            query_type_props["temporary"] = True

        query_type = _get_create_type_from_kind(kind)
        return query_type, query_type_props

    # UPGRADE: Once we use Python 3.10, replace this with a match expression.
    mapping = {
        sqlglot.exp.Select: QueryType.SELECT,
        sqlglot.exp.Insert: QueryType.INSERT,
        sqlglot.exp.Update: QueryType.UPDATE,
        sqlglot.exp.Delete: QueryType.DELETE,
        sqlglot.exp.Merge: QueryType.MERGE,
        sqlglot.exp.Query: QueryType.SELECT,  # unions, etc. are also selects
    }

    for cls, query_type in mapping.items():
        if isinstance(expression, cls):
            return query_type, query_type_props
    return QueryType.UNKNOWN, {}


def is_create_table_ddl(statement: sqlglot.exp.Expression) -> bool:
    return isinstance(statement, sqlglot.exp.Create) and isinstance(
        statement.this, sqlglot.exp.Schema
    )



import copy
import re
from typing import Any, Dict, List, Optional, cast

from datahub.configuration.common import ConfigModel
from datahub.emitter.mce_builder import Aspect
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.transformer.dataset_transformer import (
    DatasetUsageStatisticsTransformer,
)
from datahub.metadata.schema_classes import DatasetUsageStatisticsClass

_USER_URN_PREFIX: str = "urn:li:corpuser:"


class PatternCleanupDatasetUsageUserConfig(ConfigModel):
    pattern_for_cleanup: List[str]


class PatternCleanupDatasetUsageUser(DatasetUsageStatisticsTransformer):
    """Transformer that clean the user URN for DatasetUsageStatistics aspect."""

    ctx: PipelineContext
    config: PatternCleanupDatasetUsageUserConfig

    def __init__(
        self,
        config: PatternCleanupDatasetUsageUserConfig,
        ctx: PipelineContext,
        **resolver_args: Dict[str, Any],
    ):
        super().__init__()
        self.config = config
        self.ctx = ctx
        self.resolver_args = resolver_args

    @classmethod
    def create(
        cls, config_dict: dict, ctx: PipelineContext
    ) -> "PatternCleanupDatasetUsageUser":
        config = PatternCleanupDatasetUsageUserConfig.parse_obj(config_dict)
        return cls(config, ctx)

    def transform_aspect(
        self, entity_urn: str, aspect_name: str, aspect: Optional[Aspect]
    ) -> Optional[Aspect]:
        in_dataset_properties_aspect: DatasetUsageStatisticsClass = cast(
            DatasetUsageStatisticsClass, aspect
        )

        if in_dataset_properties_aspect.userCounts is not None:
            out_dataset_properties_aspect: DatasetUsageStatisticsClass = copy.deepcopy(
                in_dataset_properties_aspect
            )

            if out_dataset_properties_aspect.userCounts is not None:
                for user in out_dataset_properties_aspect.userCounts:
                    user_id: str = user.user.split(_USER_URN_PREFIX)[1]
                    for value in self.config.pattern_for_cleanup:
                        cleaned_user_id = re.sub(value, "", user_id)
                        user.user = _USER_URN_PREFIX + cleaned_user_id

                return cast(Aspect, out_dataset_properties_aspect)
            else:
                return cast(Aspect, out_dataset_properties_aspect)
        else:
            return cast(Aspect, in_dataset_properties_aspect)



from typing import Dict, List, Optional

from datahub.ingestion.graph.client import DataHubGraph


class DataHubSecretsClient:
    """Class used to fetch secrets from DataHub."""

    graph: DataHubGraph

    def __init__(self, graph: DataHubGraph):
        self.graph = graph

    def get_secret_values(self, secret_names: List[str]) -> Dict[str, Optional[str]]:
        if len(secret_names) == 0:
            return {}

        request_json = {
            "query": """query getSecretValues($input: GetSecretValuesInput!) {\n
                getSecretValues(input: $input) {\n
                    name\n
                    value\n
                }\n
            }""",
            "variables": {"input": {"secrets": secret_names}},
        }
        # TODO: Use graph.execute_graphql() instead.

        # Fetch secrets using GraphQL API f
        response = self.graph._session.post(
            f"{self.graph.config.server}/api/graphql", json=request_json
        )
        response.raise_for_status()

        # Verify response
        res_data = response.json()
        if "errors" in res_data:
            raise Exception("Failed to retrieve secrets from DataHub.")

        # Convert list of name, value secret pairs into a dict and return
        secret_value_list = res_data["data"]["getSecretValues"]
        secret_value_dict = dict()
        for secret_value in secret_value_list:
            secret_value_dict[secret_value["name"]] = secret_value["value"]
        return secret_value_dict



from datahub.ingestion.api.pipeline_run_listener import PipelineRunListener
from datahub.ingestion.api.registry import PluginRegistry

reporting_provider_registry = PluginRegistry[PipelineRunListener]()
reporting_provider_registry.register_from_entrypoint(
    "datahub.ingestion.reporting_provider.plugins"
)



import logging
import pathlib
import re
from abc import ABC, abstractmethod
from typing import Any, ClassVar, Dict, List, Optional, Set, Union

from deepmerge import always_merger
from liquid import Undefined
from liquid.exceptions import LiquidSyntaxError

from datahub.ingestion.source.looker.lkml_patched import load_lkml
from datahub.ingestion.source.looker.looker_constant import (
    DATAHUB_TRANSFORMED_SQL,
    DATAHUB_TRANSFORMED_SQL_TABLE_NAME,
    DERIVED_DOT_SQL,
    DERIVED_TABLE,
    NAME,
    SQL,
    SQL_TABLE_NAME,
    dev,
    prod,
)
from datahub.ingestion.source.looker.looker_liquid_tag import (
    CustomTagException,
    create_template,
)
from datahub.ingestion.source.looker.lookml_config import (
    DERIVED_VIEW_PATTERN,
    LookMLSourceConfig,
)

logger = logging.getLogger(__name__)


class SpecialVariable:
    SPECIAL_VARIABLE_PATTERN: ClassVar[
        str
    ] = r"\b\w+(\.\w+)*\._(is_selected|in_query|is_filtered)\b"
    liquid_variable: dict

    def __init__(self, liquid_variable):
        self.liquid_variable = liquid_variable

    def _create_new_liquid_variables_with_default(
        self,
        variables: Set[str],
    ) -> dict:
        new_dict = {**self.liquid_variable}

        for variable in variables:
            keys = variable.split(
                "."
            )  # variable is defined as view._is_selected or view.field_name._is_selected

            current_dict: dict = new_dict

            for key in keys[:-1]:
                if key not in current_dict:
                    current_dict[key] = {}

                current_dict = current_dict[key]

            if keys[-1] not in current_dict:
                current_dict[keys[-1]] = True

        logger.debug("added special variables in liquid_variable dictionary")

        return new_dict

    def liquid_variable_with_default(self, text: str) -> dict:
        variables: Set[str] = set(
            [
                text[m.start() : m.end()]
                for m in re.finditer(SpecialVariable.SPECIAL_VARIABLE_PATTERN, text)
            ]
        )

        # if set is empty then no special variables are found.
        if not variables:
            return self.liquid_variable

        return self._create_new_liquid_variables_with_default(variables=variables)


def resolve_liquid_variable(text: str, liquid_variable: Dict[Any, Any]) -> str:
    # Set variable value to NULL if not present in liquid_variable dictionary
    Undefined.__str__ = lambda instance: "NULL"  # type: ignore
    try:
        # See is there any special boolean variables are there in the text like _in_query, _is_selected, and
        # _is_filtered. Refer doc for more information
        # https://cloud.google.com/looker/docs/liquid-variable-reference#usage_of_in_query_is_selected_and_is_filtered
        # update in liquid_variable with there default values
        liquid_variable = SpecialVariable(liquid_variable).liquid_variable_with_default(
            text
        )
        # Resolve liquid template
        return create_template(text).render(liquid_variable)
    except LiquidSyntaxError as e:
        logger.warning(f"Unsupported liquid template encountered. error [{e.message}]")
        # TODO: There are some tag specific to looker and python-liquid library does not understand them. currently
        #  we are not parsing such liquid template.
        #
        # See doc: https://cloud.google.com/looker/docs/templated-filters and look for { % condition region %}
        # order.region { % endcondition %}
    except CustomTagException as e:
        logger.warning(e)
        logger.debug(e, exc_info=e)

    return text


class LookMLViewTransformer(ABC):
    """
    There are many transformations that we need to perform on the LookML view to make it suitable for metadata ingestion.

    These transformations include:

    1. Evaluating Looker templates, such as `-- if prod --` comments. Example `LookMlIfCommentTransformer`.

    2. Resolving Liquid templates. Example `LiquidVariableTransformer`.

    3. Removing ${} from derived view patterns. Example `DropDerivedViewPatternTransformer`.
       (e.g., changing ${view_name.SQL_TABLE_NAME} to 4. view_name.SQL_TABLE_NAME).

    4. Completing incomplete SQL fragments. Example `IncompleteSqlTransformer`.

    Each transformer works on specific attributes of the LookML view. For example, the #4 transformation is only
    applicable to the view.derived.sql attribute, while the other transformations apply to both the
    view.sql_table_name and view.derived.sql attributes.

    This class contains the logic to ensure that the transformer is applied to specific attributes and returns a
    dictionary containing the transformed data.

    For example:
    In case of #1 and #2, it returns:

        **transformed derived_table:**
        ```
        {
            "derived_table": {
                "datahub_transformed_sql": "<transformed value of derived_table.sql attribute>"
            }
        }
        ```

        **Whereas original was:**
        ```
        {
            "derived_table": {
                "sql": "<Sql text with liquid or lookml template language>"
            }
        }
        ```

    In case #3, it returns:
        **transformed sql_table_name:**
        ```
        {
            "datahub_transformed_sql_table_name": "employee_income_source.SQL_TABLE_NAME"
        }
        ```

        **Whereas original was:**
        ```
        {
            "sql_table_name": "${employee_income_source.SQL_TABLE_NAME}"
        }
        ```

    In case #4, it returns:
        **transformed derived_table:**
        ```
        {
            "derived_table": {
                "datahub_transformed_sql": "SELECT column_a, column_b FROM foo"
            }
        }
        ```

        **Whereas original was:**
        ```
        {
            "derived_table": {
                "sql": "column_a, column_b"
            }
        }
        ```

    Each transformation generates a section of the transformed dictionary with a new attribute named
    `datahub_transformed_<original-attribute-name>`.
    """

    source_config: LookMLSourceConfig

    def __init__(self, source_config: LookMLSourceConfig):
        self.source_config = source_config

    def transform(self, view: dict) -> dict:
        value_to_transform: Optional[str] = None

        # is_attribute_supported check is required because not all transformer works on all attributes in current
        # case mostly all transformer works on sql_table_name and derived.sql attributes,
        # however IncompleteSqlTransformer only transform the derived.sql attribute
        if SQL_TABLE_NAME in view and self.is_attribute_supported(SQL_TABLE_NAME):
            # Give precedence to already processed transformed view.sql_table_name to apply more transformation
            value_to_transform = view.get(
                DATAHUB_TRANSFORMED_SQL_TABLE_NAME, view[SQL_TABLE_NAME]
            )

        if (
            DERIVED_TABLE in view
            and SQL in view[DERIVED_TABLE]
            and self.is_attribute_supported(DERIVED_DOT_SQL)
        ):
            # Give precedence to already processed transformed view.derived.sql to apply more transformation
            value_to_transform = view[DERIVED_TABLE].get(
                DATAHUB_TRANSFORMED_SQL, view[DERIVED_TABLE][SQL]
            )

        if value_to_transform is None:
            return {}

        logger.debug(f"value to transform = {value_to_transform}")

        transformed_value: str = self._apply_transformation(
            value=value_to_transform, view=view
        )

        logger.debug(f"transformed value = {transformed_value}")

        if SQL_TABLE_NAME in view and value_to_transform:
            return {DATAHUB_TRANSFORMED_SQL_TABLE_NAME: transformed_value}

        if DERIVED_TABLE in view and SQL in view[DERIVED_TABLE] and value_to_transform:
            return {DERIVED_TABLE: {DATAHUB_TRANSFORMED_SQL: transformed_value}}

        return {}

    @abstractmethod
    def _apply_transformation(self, value: str, view: dict) -> str:
        pass

    def is_attribute_supported(self, attribute: str) -> bool:
        return attribute in [DERIVED_DOT_SQL, SQL_TABLE_NAME]


class LiquidVariableTransformer(LookMLViewTransformer):
    """
    Replace the liquid variables with their values.
    """

    def _apply_transformation(self, value: str, view: dict) -> str:
        return resolve_liquid_variable(
            text=value,
            liquid_variable=self.source_config.liquid_variable,
        )


class IncompleteSqlTransformer(LookMLViewTransformer):
    """
    lookml view may contain the fragment of sql, however for lineage generation we need a complete sql.
    IncompleteSqlTransformer will complete the view's derived.sql.
    """

    def is_attribute_supported(self, attribute: str) -> bool:
        return attribute in [DERIVED_DOT_SQL]

    def _apply_transformation(self, value: str, view: dict) -> str:
        if DERIVED_TABLE not in view or SQL not in view[DERIVED_TABLE]:
            # This transformation is only applicable in-case of view contains view.derived.sql
            return value

        # Looker supports sql fragments that omit the SELECT and FROM parts of the query
        # Add those in if we detect that it is missing
        sql_query: str = value

        if not re.search(r"SELECT\s", sql_query, flags=re.I):
            # add a SELECT clause at the beginning
            sql_query = f"SELECT {sql_query}"

        if not re.search(r"FROM\s", sql_query, flags=re.I):
            # add a FROM clause at the end
            sql_query = f"{sql_query} FROM {view[NAME]}"

        return sql_query


class DropDerivedViewPatternTransformer(LookMLViewTransformer):
    """
    drop ${} from datahub_transformed_sql_table_name and  view["derived_table"]["datahub_transformed_sql_table_name"] values.

    Example: transform ${employee_income_source.SQL_TABLE_NAME} to employee_income_source.SQL_TABLE_NAME
    """

    def _apply_transformation(self, value: str, view: dict) -> str:
        return re.sub(
            DERIVED_VIEW_PATTERN,
            r"\1",
            value,
        )


class LookMlIfCommentTransformer(LookMLViewTransformer):
    """
    Evaluate the looker -- if -- comments.
    """

    evaluate_to_true_regx: str
    remove_if_comment_line_regx: str

    def __init__(self, source_config: LookMLSourceConfig):
        super().__init__(source_config=source_config)

        # This regx will keep whatever after -- if looker_environment --
        self.evaluate_to_true_regx = r"-- if {} --".format(
            self.source_config.looker_environment
        )

        # It will remove all other lines starts with -- if ... --
        self.remove_if_comment_line_regx = r"-- if {} --.*?(?=\n|-- if|$)".format(
            dev if self.source_config.looker_environment.lower() == prod else prod
        )

    def _apply_regx(self, value: str) -> str:
        result: str = re.sub(
            self.remove_if_comment_line_regx, "", value, flags=re.IGNORECASE | re.DOTALL
        )

        # Remove '-- if prod --' but keep the rest of the line
        result = re.sub(self.evaluate_to_true_regx, "", result, flags=re.IGNORECASE)

        return result

    def _apply_transformation(self, value: str, view: dict) -> str:
        return self._apply_regx(value)


class TransformedLookMlView:
    """
    TransformedLookMlView is collecting output of LookMLViewTransformer and creating a new transformed LookML view.
    TransformedLookMlView creates a copy of the original view dictionary and updates the copy with the transformed output.
    The deepmerge library is used because Python's dict.update function doesn't merge nested fields.

    The transformed LookML view will contain the following attributes:

    ```
    {
        "derived_table": {
            "sql": "<original sql with looker template language",
            "datahub_transformed_sql": "<transformed sql>"
        },

        dimensions .....
    }
    ```
    see documentation of LookMLViewTransformer for output of each transformer.
    """

    transformers: List[LookMLViewTransformer]
    view_dict: dict
    transformed_dict: dict

    def __init__(
        self,
        transformers: List[LookMLViewTransformer],
        view_dict: dict,
    ):
        self.transformers = transformers
        self.view_dict = view_dict
        self.transformed_dict = {}

    def view(self) -> dict:
        if self.transformed_dict:
            return self.transformed_dict

        self.transformed_dict = {**self.view_dict}

        logger.debug(f"Processing view {self.view_dict[NAME]}")

        for transformer in self.transformers:
            logger.debug(f"Applying transformer {transformer.__class__.__name__}")

            self.transformed_dict = always_merger.merge(
                self.transformed_dict, transformer.transform(self.transformed_dict)
            )

        return self.transformed_dict


def process_lookml_template_language(
    source_config: LookMLSourceConfig,
    view_lkml_file_dict: dict,
) -> None:
    if "views" not in view_lkml_file_dict:
        return

    transformers: List[LookMLViewTransformer] = [
        LookMlIfCommentTransformer(
            source_config=source_config
        ),  # First evaluate the -- if -- comments. Looker does the same
        LiquidVariableTransformer(
            source_config=source_config
        ),  # Now resolve liquid variables
        DropDerivedViewPatternTransformer(
            source_config=source_config
        ),  # Remove any ${} symbol
        IncompleteSqlTransformer(
            source_config=source_config
        ),  # complete any incomplete sql
    ]

    transformed_views: List[dict] = []

    for view in view_lkml_file_dict["views"]:
        transformed_views.append(
            TransformedLookMlView(transformers=transformers, view_dict=view).view()
        )

    view_lkml_file_dict["views"] = transformed_views


def load_and_preprocess_file(
    path: Union[str, pathlib.Path],
    source_config: LookMLSourceConfig,
) -> dict:
    parsed = load_lkml(path)

    process_lookml_template_language(
        view_lkml_file_dict=parsed,
        source_config=source_config,
    )

    return parsed



import logging
from contextlib import contextmanager
from enum import Enum
from pathlib import Path
from typing import Generator, List, Optional

import yaml
from pydantic import validator
from ruamel.yaml import YAML

from datahub.configuration.common import ConfigModel
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.graph.client import DataHubGraph, get_default_graph
from datahub.metadata.schema_classes import (
    PropertyValueClass,
    StructuredPropertyDefinitionClass,
)
from datahub.utilities.urns.urn import Urn

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class StructuredPropertiesConfig:
    """Configuration class to hold the graph client"""

    _graph: Optional[DataHubGraph] = None

    @classmethod
    @contextmanager
    def use_graph(cls, graph: DataHubGraph) -> Generator[None, None, None]:
        """Context manager to temporarily set a custom graph"""
        previous_graph = cls._graph
        cls._graph = graph
        try:
            yield
        finally:
            cls._graph = previous_graph

    @classmethod
    def get_graph(cls) -> DataHubGraph:
        """Get the current graph, falling back to default if none set"""
        return cls._graph if cls._graph is not None else get_default_graph()


class AllowedTypes(Enum):
    STRING = "string"
    RICH_TEXT = "rich_text"
    NUMBER = "number"
    DATE = "date"
    URN = "urn"

    @staticmethod
    def check_allowed_type(value: str) -> bool:
        return value in [allowed_type.value for allowed_type in AllowedTypes]

    @staticmethod
    def values():
        return ", ".join([allowed_type.value for allowed_type in AllowedTypes])


class AllowedValue(ConfigModel):
    value: str
    description: Optional[str] = None


VALID_ENTITY_TYPES_PREFIX_STRING = ", ".join(
    [
        f"urn:li:entityType:datahub.{x}"
        for x in ["dataset", "dashboard", "dataFlow", "schemaField"]
    ]
)
VALID_ENTITY_TYPES_STRING = f"Valid entity type urns are {VALID_ENTITY_TYPES_PREFIX_STRING}, etc... Ensure that the entity type is valid."


class TypeQualifierAllowedTypes(ConfigModel):
    allowed_types: List[str]

    @validator("allowed_types", each_item=True)
    def validate_allowed_types(cls, v):
        if v:
            graph = StructuredPropertiesConfig.get_graph()
            validated_urn = Urn.make_entity_type_urn(v)
            if not graph.exists(validated_urn):
                raise ValueError(
                    f"Input {v} is not a valid entity type urn. {VALID_ENTITY_TYPES_STRING}"
                )
            v = str(validated_urn)
        return v


class StructuredProperties(ConfigModel):
    id: Optional[str] = None
    urn: Optional[str] = None
    qualified_name: Optional[str] = None
    type: str
    value_entity_types: Optional[List[str]] = None
    description: Optional[str] = None
    display_name: Optional[str] = None
    entity_types: Optional[List[str]] = None
    cardinality: Optional[str] = None
    allowed_values: Optional[List[AllowedValue]] = None
    type_qualifier: Optional[TypeQualifierAllowedTypes] = None
    immutable: Optional[bool] = False

    @validator("entity_types", each_item=True)
    def validate_entity_types(cls, v):
        if v:
            graph = StructuredPropertiesConfig.get_graph()
            validated_urn = Urn.make_entity_type_urn(v)
            if not graph.exists(validated_urn):
                raise ValueError(
                    f"Input {v} is not a valid entity type urn. {VALID_ENTITY_TYPES_STRING}"
                )
            v = str(validated_urn)
        return v

    @property
    def fqn(self) -> str:
        assert self.urn is not None
        return (
            self.qualified_name
            or self.id
            or Urn.create_from_string(self.urn).get_entity_id()[0]
        )

    @validator("urn", pre=True, always=True)
    def urn_must_be_present(cls, v, values):
        if not v:
            if "id" not in values:
                raise ValueError("id must be present if urn is not")
            return f"urn:li:structuredProperty:{values['id']}"
        return v

    @staticmethod
    def create(file: str, graph: Optional[DataHubGraph] = None) -> None:
        emitter: DataHubGraph = graph if graph else get_default_graph()
        with StructuredPropertiesConfig.use_graph(emitter):
            print("Using graph")
            with open(file) as fp:
                structuredproperties: List[dict] = yaml.safe_load(fp)
                for structuredproperty_raw in structuredproperties:
                    structuredproperty = StructuredProperties.parse_obj(
                        structuredproperty_raw
                    )
                    if not structuredproperty.type.islower():
                        structuredproperty.type = structuredproperty.type.lower()
                        logger.warn(
                            f"Structured property type should be lowercase. Updated to {structuredproperty.type}"
                        )
                    if not AllowedTypes.check_allowed_type(structuredproperty.type):
                        raise ValueError(
                            f"Type {structuredproperty.type} is not allowed. Allowed types are {AllowedTypes.values()}"
                        )
                    mcp = MetadataChangeProposalWrapper(
                        entityUrn=structuredproperty.urn,
                        aspect=StructuredPropertyDefinitionClass(
                            qualifiedName=structuredproperty.fqn,
                            valueType=Urn.make_data_type_urn(structuredproperty.type),
                            displayName=structuredproperty.display_name,
                            description=structuredproperty.description,
                            entityTypes=[
                                Urn.make_entity_type_urn(entity_type)
                                for entity_type in structuredproperty.entity_types or []
                            ],
                            cardinality=structuredproperty.cardinality,
                            immutable=structuredproperty.immutable,
                            allowedValues=(
                                [
                                    PropertyValueClass(
                                        value=v.value, description=v.description
                                    )
                                    for v in structuredproperty.allowed_values
                                ]
                                if structuredproperty.allowed_values
                                else None
                            ),
                            typeQualifier=(
                                {
                                    "allowedTypes": structuredproperty.type_qualifier.allowed_types
                                }
                                if structuredproperty.type_qualifier
                                else None
                            ),
                        ),
                    )
                    emitter.emit_mcp(mcp)

                    logger.info(f"Created structured property {structuredproperty.urn}")

    @classmethod
    def from_datahub(cls, graph: DataHubGraph, urn: str) -> "StructuredProperties":
        with StructuredPropertiesConfig.use_graph(graph):
            structured_property: Optional[
                StructuredPropertyDefinitionClass
            ] = graph.get_aspect(urn, StructuredPropertyDefinitionClass)
            if structured_property is None:
                raise Exception(
                    "StructuredPropertyDefinition aspect is None. Unable to create structured property."
                )
            return StructuredProperties(
                urn=urn,
                qualified_name=structured_property.qualifiedName,
                display_name=structured_property.displayName,
                type=structured_property.valueType,
                description=structured_property.description,
                entity_types=structured_property.entityTypes,
                cardinality=structured_property.cardinality,
                allowed_values=(
                    [
                        AllowedValue(
                            value=av.value,
                            description=av.description,
                        )
                        for av in structured_property.allowedValues or []
                    ]
                    if structured_property.allowedValues is not None
                    else None
                ),
                type_qualifier=(
                    {
                        "allowed_types": structured_property.typeQualifier.get(
                            "allowedTypes"
                        )
                    }
                    if structured_property.typeQualifier
                    else None
                ),
            )

    def to_yaml(
        self,
        file: Path,
    ) -> None:
        with open(file, "w") as fp:
            yaml = YAML(typ="rt")  # default, if not specfied, is 'rt' (round-trip)
            yaml.indent(mapping=2, sequence=4, offset=2)
            yaml.default_flow_style = False
            yaml.dump(self.dict(), fp)



import logging
from typing import Callable, Dict, Iterable, List, Optional, Tuple, Type, Union, cast

from avrogen.dict_wrapper import DictWrapper
from pydantic import BaseModel

import datahub.metadata.schema_classes as models
from datahub.api.entities.common.data_platform_instance import DataPlatformInstance
from datahub.api.entities.common.serialized_value import SerializedResourceValue
from datahub.emitter.mce_builder import (
    make_data_platform_urn,
    make_dataplatform_instance_urn,
)
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.emitter.mcp_builder import DatahubKey
from datahub.ingestion.graph.client import DataHubGraph
from datahub.metadata.urns import (
    DataPlatformInstanceUrn,
    DataPlatformUrn,
    PlatformResourceUrn,
    Urn,
)
from datahub.utilities.openapi_utils import OpenAPIGraphClient
from datahub.utilities.search_utils import (
    ElasticDocumentQuery,
    ElasticsearchQueryBuilder,
    LogicalOperator,
    SearchField,
)

logger = logging.getLogger(__name__)


class PlatformResourceKey(DatahubKey):
    platform: str
    platform_instance: Optional[str] = None
    resource_type: str
    primary_key: str

    @property
    def id(self) -> str:
        return self.guid()


class PlatformResourceInfo(BaseModel):
    resource_type: str
    primary_key: str
    value: Optional[SerializedResourceValue] = None
    secondary_keys: Optional[List[str]] = None

    @classmethod
    def from_resource_info(
        cls, resource_info: models.PlatformResourceInfoClass
    ) -> "PlatformResourceInfo":
        serialized_value: Optional[SerializedResourceValue] = None
        if resource_info.value:
            serialized_value = SerializedResourceValue.from_resource_value(
                resource_info.value
            )
        return cls(
            primary_key=resource_info.primaryKey,
            secondary_keys=resource_info.secondaryKeys,
            resource_type=resource_info.resourceType,
            value=serialized_value,
        )

    def to_resource_info(self) -> models.PlatformResourceInfoClass:
        value = None
        if self.value:
            value = models.SerializedValueClass(
                contentType=self.value.content_type,
                blob=self.value.blob,
                schemaType=self.value.schema_type,
                schemaRef=self.value.schema_ref,
            )
        return models.PlatformResourceInfoClass(
            primaryKey=self.primary_key,
            secondaryKeys=self.secondary_keys,
            resourceType=self.resource_type,
            value=value,
        )


class UrnSearchField(SearchField):
    """
    A search field that supports URN values.
    TODO: Move this to search_utils after we make this more generic.
    """

    def __init__(self, field_name: str, urn_value_extractor: Callable[[str], Urn]):
        self.urn_value_extractor = urn_value_extractor
        super().__init__(field_name)

    def get_search_value(self, value: str) -> str:
        return str(self.urn_value_extractor(value))


class PlatformResourceSearchField(SearchField):
    def __init__(self, field_name: str):
        super().__init__(field_name)

    @classmethod
    def from_search_field(
        cls, search_field: SearchField
    ) -> "PlatformResourceSearchField":
        # pretends to be a class method, but just returns the input
        return search_field  # type: ignore


class PlatformResourceSearchFields:
    PRIMARY_KEY = PlatformResourceSearchField("primaryKey")
    RESOURCE_TYPE = PlatformResourceSearchField("resourceType")
    SECONDARY_KEYS = PlatformResourceSearchField("secondaryKeys")
    PLATFORM = PlatformResourceSearchField.from_search_field(
        UrnSearchField(
            field_name="platform.keyword",
            urn_value_extractor=DataPlatformUrn.create_from_id,
        )
    )
    PLATFORM_INSTANCE = PlatformResourceSearchField.from_search_field(
        UrnSearchField(
            field_name="platformInstance.keyword",
            urn_value_extractor=DataPlatformInstanceUrn.from_string,
        )
    )


class ElasticPlatformResourceQuery(ElasticDocumentQuery[PlatformResourceSearchField]):
    def __init__(self):
        super().__init__()

    @classmethod
    def create_from(
        cls: Type["ElasticPlatformResourceQuery"],
        *args: Tuple[Union[str, PlatformResourceSearchField], str],
    ) -> "ElasticPlatformResourceQuery":
        return cast(ElasticPlatformResourceQuery, super().create_from(*args))


class PlatformResource(BaseModel):
    id: str
    resource_info: Optional[PlatformResourceInfo] = None
    data_platform_instance: Optional[DataPlatformInstance] = None
    removed: bool = False

    @classmethod
    def remove(
        cls,
        key: PlatformResourceKey,
    ) -> "PlatformResource":
        """
        Creates a PlatformResource object with the removed status set to True.
        Removed PlatformResource objects are used to soft-delete resources from
        the graph.
        To hard-delete a resource, use the delete method.
        """
        return cls(
            id=key.id,
            removed=True,
        )

    @classmethod
    def create(
        cls,
        key: PlatformResourceKey,
        value: Union[Dict, DictWrapper, BaseModel],
        secondary_keys: Optional[List[str]] = None,
    ) -> "PlatformResource":
        return cls(
            id=key.id,
            resource_info=PlatformResourceInfo(
                resource_type=key.resource_type,
                primary_key=key.primary_key,
                secondary_keys=secondary_keys,
                value=SerializedResourceValue.create(value),
            ),
            removed=False,
            data_platform_instance=DataPlatformInstance(
                platform=make_data_platform_urn(key.platform),
                platform_instance=(
                    make_dataplatform_instance_urn(
                        platform=key.platform, instance=key.platform_instance
                    )
                    if key.platform_instance
                    else None
                ),
            ),
        )

    def to_mcps(self) -> Iterable[MetadataChangeProposalWrapper]:
        dpi = (
            self.data_platform_instance.to_data_platform_instance()
            if self.data_platform_instance
            else None
        )
        yield from MetadataChangeProposalWrapper.construct_many(
            entityUrn=str(PlatformResourceUrn(self.id)),
            aspects=[
                self.resource_info.to_resource_info() if self.resource_info else None,
                dpi,
                models.StatusClass(removed=self.removed),
            ],
        )

    def to_datahub(self, graph_client: DataHubGraph) -> None:
        for mcp in self.to_mcps():
            graph_client.emit(mcp)

    @classmethod
    def from_datahub(
        cls, graph_client: DataHubGraph, key: Union[PlatformResourceKey, str]
    ) -> Optional["PlatformResource"]:
        """
        Fetches a PlatformResource from the graph given a key.
        Key can be either a PlatformResourceKey object or an urn string.
        Returns None if the resource is not found.
        """
        if isinstance(key, PlatformResourceKey):
            urn = PlatformResourceUrn(id=key.id)
        else:
            urn = PlatformResourceUrn.from_string(key)
        if not graph_client.exists(str(urn)):
            return None
        platform_resource = graph_client.get_entity_semityped(str(urn))
        return cls(
            id=urn.id,
            resource_info=(
                PlatformResourceInfo.from_resource_info(
                    platform_resource["platformResourceInfo"]
                )
                if "platformResourceInfo" in platform_resource
                else None
            ),
            data_platform_instance=(
                DataPlatformInstance.from_data_platform_instance(
                    platform_resource["dataPlatformInstance"]
                )
                if "dataPlatformInstance" in platform_resource
                else None
            ),
            removed=(
                platform_resource["status"].removed
                if "status" in platform_resource
                else False
            ),
        )

    @staticmethod
    def search_by_key(
        graph_client: DataHubGraph,
        key: str,
        primary: bool = True,
        is_exact: bool = True,
    ) -> Iterable["PlatformResource"]:
        """
        Searches for PlatformResource entities by primary or secondary key.

        :param graph_client: DataHubGraph client
        :param key: The key to search for
        :param primary: Whether to search for primary only or expand the search
            to secondary keys (default: True)
        :param is_exact: Whether to search for an exact match (default: True)
        :return: An iterable of PlatformResource objects
        """

        elastic_platform_resource_group = (
            ElasticPlatformResourceQuery.create_from()
            .group(LogicalOperator.OR)
            .add_field_match(
                PlatformResourceSearchFields.PRIMARY_KEY, key, is_exact=is_exact
            )
        )
        if not primary:  # we expand the search to secondary keys
            elastic_platform_resource_group.add_field_match(
                PlatformResourceSearchFields.SECONDARY_KEYS, key, is_exact=is_exact
            )
        query = elastic_platform_resource_group.end()
        openapi_client = OpenAPIGraphClient(graph_client)
        for urn in openapi_client.scroll_urns_by_filter(
            entity_type="platformResource",
            query=query,
        ):
            platform_resource = PlatformResource.from_datahub(graph_client, urn)
            if platform_resource:
                yield platform_resource

    def delete(self, graph_client: DataHubGraph, hard: bool = True) -> None:
        graph_client.delete_entity(str(PlatformResourceUrn(self.id)), hard=hard)

    @staticmethod
    def search_by_filters(
        graph_client: DataHubGraph,
        query: Union[
            ElasticPlatformResourceQuery,
            ElasticDocumentQuery,
            ElasticsearchQueryBuilder,
        ],
    ) -> Iterable["PlatformResource"]:
        openapi_client = OpenAPIGraphClient(graph_client)
        for urn in openapi_client.scroll_urns_by_filter(
            entity_type="platformResource",
            query=query,
        ):
            platform_resource = PlatformResource.from_datahub(graph_client, urn)
            if platform_resource:
                yield platform_resource



from typing import Dict, Generic, List, Optional, Tuple, TypeVar, Union

from datahub.emitter.mcp_patch_builder import MetadataPatchProposal
from datahub.metadata.com.linkedin.pegasus2avro.common import TimeStamp
from datahub.metadata.schema_classes import (
    DatasetPropertiesClass as DatasetProperties,
    EditableDatasetPropertiesClass as EditableDatasetProperties,
    EditableSchemaMetadataClass as EditableSchemaMetadata,
    FineGrainedLineageClass as FineGrainedLineage,
    GlobalTagsClass as GlobalTags,
    GlossaryTermAssociationClass as Term,
    GlossaryTermsClass as GlossaryTerms,
    KafkaAuditHeaderClass,
    OwnerClass as Owner,
    OwnershipTypeClass,
    SchemaMetadataClass,
    SystemMetadataClass,
    TagAssociationClass as Tag,
    UpstreamClass as Upstream,
    UpstreamLineageClass as UpstreamLineage,
)
from datahub.specific.custom_properties import CustomPropertiesPatchHelper
from datahub.specific.ownership import OwnershipPatchHelper
from datahub.specific.structured_properties import StructuredPropertiesPatchHelper
from datahub.utilities.urns.tag_urn import TagUrn
from datahub.utilities.urns.urn import Urn

_Parent = TypeVar("_Parent", bound=MetadataPatchProposal)


class FieldPatchHelper(Generic[_Parent]):
    def __init__(
        self,
        parent: _Parent,
        field_path: str,
        editable: bool = True,
    ) -> None:
        self._parent: _Parent = parent
        self.field_path = field_path
        self.aspect_name = (
            EditableSchemaMetadata.ASPECT_NAME
            if editable
            else SchemaMetadataClass.ASPECT_NAME
        )
        self.aspect_field = "editableSchemaFieldInfo" if editable else "schemaFieldInfo"

    def add_tag(self, tag: Tag) -> "FieldPatchHelper":
        self._parent._add_patch(
            self.aspect_name,
            "add",
            path=f"/{self.aspect_field}/{self.field_path}/globalTags/tags/{tag.tag}",
            value=tag,
        )
        return self

    def remove_tag(self, tag: Union[str, Urn]) -> "FieldPatchHelper":
        if isinstance(tag, str) and not tag.startswith("urn:li:tag:"):
            tag = TagUrn.create_from_id(tag)
        self._parent._add_patch(
            self.aspect_name,
            "remove",
            path=f"/{self.aspect_field}/{self.field_path}/globalTags/tags/{tag}",
            value={},
        )
        return self

    def add_term(self, term: Term) -> "FieldPatchHelper":
        self._parent._add_patch(
            self.aspect_name,
            "add",
            path=f"/{self.aspect_field}/{self.field_path}/glossaryTerms/terms/{term.urn}",
            value=term,
        )
        return self

    def remove_term(self, term: Union[str, Urn]) -> "FieldPatchHelper":
        if isinstance(term, str) and not term.startswith("urn:li:glossaryTerm:"):
            term = "urn:li:glossaryTerm:" + term
        self._parent._add_patch(
            self.aspect_name,
            "remove",
            path=f"/{self.aspect_field}/{self.field_path}/glossaryTerms/terms/{term}",
            value={},
        )
        return self

    def parent(self) -> _Parent:
        return self._parent


class DatasetPatchBuilder(MetadataPatchProposal):
    def __init__(
        self,
        urn: str,
        system_metadata: Optional[SystemMetadataClass] = None,
        audit_header: Optional[KafkaAuditHeaderClass] = None,
    ) -> None:
        super().__init__(
            urn, system_metadata=system_metadata, audit_header=audit_header
        )
        self.custom_properties_patch_helper = CustomPropertiesPatchHelper(
            self, DatasetProperties.ASPECT_NAME
        )
        self.ownership_patch_helper = OwnershipPatchHelper(self)
        self.structured_properties_patch_helper = StructuredPropertiesPatchHelper(self)

    def add_owner(self, owner: Owner) -> "DatasetPatchBuilder":
        self.ownership_patch_helper.add_owner(owner)
        return self

    def remove_owner(
        self, owner: str, owner_type: Optional[OwnershipTypeClass] = None
    ) -> "DatasetPatchBuilder":
        """
        param: owner_type is optional
        """
        self.ownership_patch_helper.remove_owner(owner, owner_type)
        return self

    def set_owners(self, owners: List[Owner]) -> "DatasetPatchBuilder":
        self.ownership_patch_helper.set_owners(owners)
        return self

    def add_upstream_lineage(self, upstream: Upstream) -> "DatasetPatchBuilder":
        self._add_patch(
            UpstreamLineage.ASPECT_NAME,
            "add",
            path=f"/upstreams/{self.quote(upstream.dataset)}",
            value=upstream,
        )
        return self

    def remove_upstream_lineage(
        self, dataset: Union[str, Urn]
    ) -> "DatasetPatchBuilder":
        self._add_patch(
            UpstreamLineage.ASPECT_NAME,
            "remove",
            path=f"/upstreams/{dataset}",
            value={},
        )
        return self

    def set_upstream_lineages(self, upstreams: List[Upstream]) -> "DatasetPatchBuilder":
        self._add_patch(
            UpstreamLineage.ASPECT_NAME, "add", path="/upstreams", value=upstreams
        )
        return self

    def add_fine_grained_upstream_lineage(
        self, fine_grained_lineage: FineGrainedLineage
    ) -> "DatasetPatchBuilder":
        (
            transform_op,
            downstream_urn,
            query_id,
        ) = DatasetPatchBuilder.get_fine_grained_key(fine_grained_lineage)
        for upstream_urn in fine_grained_lineage.upstreams or []:
            self._add_patch(
                UpstreamLineage.ASPECT_NAME,
                "add",
                path=DatasetPatchBuilder.quote_fine_grained_path(
                    transform_op, downstream_urn, query_id, upstream_urn
                ),
                value={"confidenceScore": fine_grained_lineage.confidenceScore},
            )
        return self

    @staticmethod
    def get_fine_grained_key(
        fine_grained_lineage: FineGrainedLineage,
    ) -> Tuple[str, str, str]:
        downstreams = fine_grained_lineage.downstreams or []
        if len(downstreams) != 1:
            raise TypeError("Cannot patch with more or less than one downstream.")
        transform_op = fine_grained_lineage.transformOperation or "NONE"
        downstream_urn = downstreams[0]
        query_id = fine_grained_lineage.query or "NONE"
        return transform_op, downstream_urn, query_id

    @classmethod
    def quote_fine_grained_path(
        cls, transform_op: str, downstream_urn: str, query_id: str, upstream_urn: str
    ) -> str:
        return (
            f"/fineGrainedLineages/{cls.quote(transform_op)}/"
            f"{cls.quote(downstream_urn)}/{cls.quote(query_id)}/{cls.quote(upstream_urn)}"
        )

    def remove_fine_grained_upstream_lineage(
        self, fine_grained_lineage: FineGrainedLineage
    ) -> "DatasetPatchBuilder":
        (
            transform_op,
            downstream_urn,
            query_id,
        ) = DatasetPatchBuilder.get_fine_grained_key(fine_grained_lineage)
        for upstream_urn in fine_grained_lineage.upstreams or []:
            self._add_patch(
                UpstreamLineage.ASPECT_NAME,
                "remove",
                path=DatasetPatchBuilder.quote_fine_grained_path(
                    transform_op, downstream_urn, query_id, upstream_urn
                ),
                value={},
            )
        return self

    def set_fine_grained_upstream_lineages(
        self, fine_grained_lineages: List[FineGrainedLineage]
    ) -> "DatasetPatchBuilder":
        self._add_patch(
            UpstreamLineage.ASPECT_NAME,
            "add",
            path="/fineGrainedLineages",
            value=fine_grained_lineages,
        )
        return self

    def add_tag(self, tag: Tag) -> "DatasetPatchBuilder":
        self._add_patch(
            GlobalTags.ASPECT_NAME, "add", path=f"/tags/{tag.tag}", value=tag
        )
        return self

    def remove_tag(self, tag: Union[str, Urn]) -> "DatasetPatchBuilder":
        if isinstance(tag, str) and not tag.startswith("urn:li:tag:"):
            tag = TagUrn.create_from_id(tag)
        self._add_patch(GlobalTags.ASPECT_NAME, "remove", path=f"/tags/{tag}", value={})
        return self

    def add_term(self, term: Term) -> "DatasetPatchBuilder":
        self._add_patch(
            GlossaryTerms.ASPECT_NAME, "add", path=f"/terms/{term.urn}", value=term
        )
        return self

    def remove_term(self, term: Union[str, Urn]) -> "DatasetPatchBuilder":
        if isinstance(term, str) and not term.startswith("urn:li:glossaryTerm:"):
            term = "urn:li:glossaryTerm:" + term
        self._add_patch(
            GlossaryTerms.ASPECT_NAME, "remove", path=f"/terms/{term}", value={}
        )
        return self

    def for_field(
        self, field_path: str, editable: bool = True
    ) -> FieldPatchHelper["DatasetPatchBuilder"]:
        """
        Get a helper that can perform patches against fields in the dataset

        :param field_path: The field path in datahub format
        :param editable: Whether patches should apply to the editable section of schema metadata or not
        """
        return FieldPatchHelper(
            self,
            field_path,
            editable=editable,
        )

    def set_description(
        self, description: Optional[str] = None, editable: bool = False
    ) -> "DatasetPatchBuilder":
        if description is not None:
            self._add_patch(
                (
                    DatasetProperties.ASPECT_NAME
                    if not editable
                    else EditableDatasetProperties.ASPECT_NAME
                ),
                "add",
                path="/description",
                value=description,
            )
        return self

    def set_custom_properties(
        self, custom_properties: Dict[str, str]
    ) -> "DatasetPatchBuilder":
        self._add_patch(
            DatasetProperties.ASPECT_NAME,
            "add",
            path="/customProperties",
            value=custom_properties,
        )
        return self

    def add_custom_property(self, key: str, value: str) -> "DatasetPatchBuilder":
        self.custom_properties_patch_helper.add_property(key, value)
        return self

    def add_custom_properties(
        self, custom_properties: Optional[Dict[str, str]] = None
    ) -> "DatasetPatchBuilder":
        if custom_properties is not None:
            for key, value in custom_properties.items():
                self.custom_properties_patch_helper.add_property(key, value)
        return self

    def remove_custom_property(self, key: str) -> "DatasetPatchBuilder":
        self.custom_properties_patch_helper.remove_property(key)
        return self

    def set_display_name(
        self, display_name: Optional[str] = None
    ) -> "DatasetPatchBuilder":
        if display_name is not None:
            self._add_patch(
                DatasetProperties.ASPECT_NAME,
                "add",
                path="/name",
                value=display_name,
            )
        return self

    def set_qualified_name(
        self, qualified_name: Optional[str] = None
    ) -> "DatasetPatchBuilder":
        if qualified_name is not None:
            self._add_patch(
                DatasetProperties.ASPECT_NAME,
                "add",
                path="/qualifiedName",
                value=qualified_name,
            )
        return self

    def set_created(
        self, timestamp: Optional[TimeStamp] = None
    ) -> "DatasetPatchBuilder":
        if timestamp is not None:
            self._add_patch(
                DatasetProperties.ASPECT_NAME,
                "add",
                path="/created",
                value=timestamp,
            )
        return self

    def set_last_modified(
        self, timestamp: Optional[TimeStamp] = None
    ) -> "DatasetPatchBuilder":
        if timestamp is not None:
            self._add_patch(
                DatasetProperties.ASPECT_NAME,
                "add",
                path="/lastModified",
                value=timestamp,
            )
        return self

    def set_structured_property(
        self, property_name: str, value: Union[str, float, List[Union[str, float]]]
    ) -> "DatasetPatchBuilder":
        """
        This is a helper method to set a structured property.
        @param property_name: the name of the property (either bare or urn form)
        @param value: the value of the property (for multi-valued properties, this can be a list)
        """
        self.structured_properties_patch_helper.set_property(property_name, value)
        return self

    def add_structured_property(
        self, property_name: str, value: Union[str, float]
    ) -> "DatasetPatchBuilder":
        """
        This is a helper method to add a structured property.
        @param property_name: the name of the property (either bare or urn form)
        @param value: the value of the property (for multi-valued properties, this value will be appended to the list)
        """
        self.structured_properties_patch_helper.add_property(property_name, value)
        return self

    def remove_structured_property(self, property_name: str) -> "DatasetPatchBuilder":
        """
        This is a helper method to remove a structured property.
        @param property_name: the name of the property (either bare or urn form)
        """
        self.structured_properties_patch_helper.remove_property(property_name)
        return self
